import numpy as np
import re

from os.path import isfile

section_re = re.compile(r'^ *\[ *([a-zA-Z0-9_.-]+) *\] *$')


class section_index:
    """Read an ini-style gromacs index file

    Reads and parses named index file, keep a list
    of name-array-tuples, containing
    name and indices of the specified (non-empty) sections.
    """
    def __init__(self, filename, max_index):

        if filename is not None and isfile(filename):
            sections = []
            members = []
            name = None
            with open(filename, 'r') as f:
                for L in f:
                    m = section_re.match(L)
                    if m:
                        if members and name:
                            sections.append(
                                (name, np.unique(np.concatenate(members))-1))
                        name = m.group(1)
                        members = []
                    elif not L.isspace():
                        members.append(np.fromstring(L, dtype=int, sep=' '))
                if members and name:
                    sections.append(
                        (name, np.unique(np.concatenate(members))-1))
        else:
            sections = [("all", np.arange(max_index, dtype=int))]

        self.sections = sections
        if not self.valid_index_limits(max_index):
            raise ValueError('section_index: Index file seems to contain one'
                             ' or more invalid indices. For the provided'
                             ' trajectory file, indices must be in range'
                             ' [1, {}]'.format(max_index))

    def valid_index_limits(self, N):
        for _, I in self.sections:
            if I[0] < 0 or I[-1] >= N:
                return False
        return True

    def get_section_names(self):
        return [n for n, _ in self.sections]

    def get_section_indices(self):
        return [i for _, i in self.sections]

    def N_sections(self):
        return len(self.sections)

    def get_section_split_function(self):
        """Special function for splitting (3,N) dimensioned x or v arrays

        Split x/v into list of xs/vs in accordance with specified sections.
        """
        indices = [I for _, I in self.sections]

        def fun(frame):
            frame = frame.copy()
            frame['xs'] = [frame['x'][:, I] for I in indices]
            if 'v' in frame:
                frame['vs'] = [frame['v'][:, I] for I in indices]
            return frame
        return fun



import numpy as np
import re

from collections import deque
from dsf.trajectory_reader.abstract_trajectory_reader \
    import abstract_trajectory_reader
from itertools import count
from numpy import array, arange, zeros


class lammpstrj_trajectory_reader(abstract_trajectory_reader):
    """Read LAMMPS trajectory file

    This is a naive (and comparatively slow) implementation,
    written entirely in python.
    """

    @classmethod
    def available(cls):
        return True

    def __init__(self, filename, x_factor=0.1, t_factor=1.0):
        if filename.endswith('.gz'):
            import gzip
            self._fh = gzip.open(filename, 'rt')
        elif filename.endswith('.bz2'):
            import bz2
            self._fh = bz2.open(filename, 'rt')
        else:
            self._fh = open(filename, 'r')

        self._open = True
        regexp = r'^ITEM: (TIMESTEP|NUMBER OF ATOMS|BOX BOUNDS|ATOMS) ?(.*)$'
        self._item_re = re.compile(regexp)
        self.x_factor = x_factor
        self.t_factor = t_factor
        self.v_factor = x_factor / t_factor
        self._first_called = False
        self._index = count(1)

    # ITEM: TIMESTEP
    # 81000
    # ITEM: NUMBER OF ATOMS
    # 1536
    # ITEM: BOX BOUNDS pp pp pp
    # 1.54223 26.5378
    # 1.54223 26.5378
    # 1.54223 26.5378
    # ITEM: ATOMS id type x y z vx vy vz
    # 247 1 3.69544 2.56202 3.27701 0.00433856 -0.00099307 -0.00486166
    # 249 2 3.73324 3.05962 4.14359 0.00346029 0.00332502 -0.00731005
    # 463 1 3.5465 4.12841 5.34888 0.000523332 0.00145597 -0.00418675

    def _read_frame_header(self):
        while True:
            L = self._fh.readline()
            m = self._item_re.match(L)
            if not m:
                if L == '':
                    self._fh.close()
                    self._open = False
                    raise StopIteration
                if L.strip() == '':
                    continue
                raise IOError('TRJ_reader: Failed to parse TRJ frame header')
            if m.group(1) == 'TIMESTEP':
                step = int(self._fh.readline())
            elif m.group(1) == 'NUMBER OF ATOMS':
                natoms = int(self._fh.readline())
            elif m.group(1) == 'BOX BOUNDS':
                bbounds = [deque(map(float, self._fh.readline().split()))
                           for _ in range(3)]
                x = array(bbounds)
                box = np.diag(x[:, 1] - x[:, 0])
                if x.shape == (3, 3):
                    box[1, 0] = x[0, 2]
                    box[2, 0] = x[1, 2]
                    box[2, 1] = x[2, 2]
                elif x.shape != (3, 2):
                    raise IOError('TRJ_reader: Malformed box bounds')
            elif m.group(1) == 'ATOMS':
                cols = tuple(m.group(2).split())
                # At this point, there should be only atomic data left
                return (step, natoms, box, cols)

    def _get_first(self):
        # Read first frame, update state of self, create indexes etc
        step, N, box, cols = self._read_frame_header()
        self._natoms = N
        self._step = step
        self._cols = cols
        self._box = box

        def _all_in_cols(keys):
            for k in keys:
                if k not in cols:
                    return False
            return True

        self._x_map = None
        if _all_in_cols(('id', 'xu', 'yu', 'zu')):
            self._x_I = array(deque(map(cols.index, ('xu', 'yu', 'zu'))))
        elif _all_in_cols(('id', 'x', 'y', 'z')):
            self._x_I = array(deque(map(cols.index, ('x', 'y', 'z'))))
        elif _all_in_cols(('id', 'xs', 'ys', 'zs')):
            self._x_I = array(deque(map(cols.index, ('xs', 'ys', 'zs'))))
            _x_factor = self._box.diagonal().reshape((3, 1))
            # xs.shape == (3,n)
            self._x_map = lambda xs: xs * _x_factor
        else:
            raise RuntimeError('TRJ file must contain at least atom-id, x, y,'
                               ' and z coordinates to be useful.')
        self._id_I = cols.index('id')

        if _all_in_cols(('vx', 'vy', 'vz')):
            self._v_I = array(deque(map(cols.index, ('vx', 'vy', 'vz'))))
        else:
            self._v_I = None

        if 'type' in cols:
            self._type_I = cols.index('type')
        else:
            self._type_I = None

        data = array([list(map(float, self._fh.readline().split())) for _ in range(N)])
        I = np.asarray(data[:, self._id_I], dtype=np.int)
        # Unless dump is done for group 'all' ...
        I[np.argsort(I)] = arange(len(I))
        self._x = zeros((3, N), order='F')
        if self._x_map is None:
            self._x[:, I] = data[:, self._x_I].transpose()
        else:
            self._x[:, I] = self._x_map(data[:, self._x_I].transpose())
        if self._v_I is not None:
            self._v = zeros((3, N), order='F')
            self._v[:, I] = data[:, self._v_I].transpose()

    def _get_next(self):
        # get next frame, update state of self
        step, N, box, cols = self._read_frame_header()
        assert(self._natoms == N)
        assert(self._cols == cols)
        self._step = step
        self._box = box

        data = array([deque(map(float, self._fh.readline().split()))
                      for _ in range(N)])
        I = np.asarray(data[:, self._id_I], dtype=np.int) - 1
        if self._x_map is None:
            self._x[:, I] = data[:, self._x_I].transpose()
        else:
            self._x[:, I] = self._x_map(data[:, self._x_I].transpose())
        if self._v_I is not None:
            self._v[:, I] = data[:, self._v_I].transpose()

    def __iter__(self):
        return self

    def close(self):
        if not self._fh.closed:
            self._fh.close()

    def __next__(self):
        if not self._open:
            raise StopIteration

        if self._first_called:
            self._get_next()
        else:
            self._get_first()

        res = dict(
            index=next(self._index),
            N=int(self._natoms),
            box=self.x_factor * self._box.copy('F'),
            time=self.t_factor * self._step,
            x=self.x_factor * self._x,
            )

        if self._v_I is not None:
            res['v'] = self.v_factor * self._v

        return res



import unittest
import dsf.trajectory_reader.molfile_plugin as mplugin


_molfile_not_found = mplugin.molfile_plugin_dir() is None

class MolfilePluginTest(unittest.TestCase):
    @unittest.skipIf(_molfile_not_found, "Molfile plugins not found")
    def test_can_load_plugins(self):
        for _, _, _, plugin_name in mplugin.TRAJECTORY_PLUGIN_MAPPING:
            plugin_path = mplugin.molfile_plugin_path(plugin_name)
            mp = mplugin._MolfilePlugin(plugin_path)
            print("%-20s %-15s (%5s %5s %5s %5s)" % \
                      (plugin_name, mp.plugin.filename_extension,
                       bool(mp.plugin.read_timestep),
                       bool(mp.plugin.read_timestep_metadata),
                       bool(mp.plugin.read_next_timestep),
                       bool(mp.plugin.read_structure)))
            mp.close()



# -*- coding: utf-8 -*-

"""
dynasor module.
"""

__project__ = 'dynasor'
__description__ = 'A tool for calculating dynamical structure factors'
__copyright__ = '2021'
__license__ = 'MIT'
__credits__ = ['The dynasor developers team']
__version__ = '1.1.1'
__maintainer__ = 'The dynasor developers team'
__maintainer_email__ = 'dynasor@materialsmodeling.org'
__status__ = 'Development Status :: 5 - Production/Stable'
__url__ = 'http://dynasor.materialsmodeling.org/'



import unittest
from dsf.trajectory_reader.lammpstrj_trajectory_reader \
    import lammpstrj_trajectory_reader as trajectory_reader
from dsf.trajectory_reader.test.trajectory_reader_test_mixin \
    import TrajectoryReaderTestMixin


_not_available = not trajectory_reader.available()
_not_available_reason = 'lammpstrj plugin not available'


class LAMMPSTRJTrajectoryReaderTest(unittest.TestCase,
                                    TrajectoryReaderTestMixin):

    @unittest.skipIf(_not_available, _not_available_reason)
    def test_open_lammpstrj_no_velocities(self):
        trajectory_reader(self.filename_lammpstrj_no_velocities())

    @unittest.skipIf(_not_available, _not_available_reason)
    def test_open_lammpstrj(self):
        trajectory_reader(self.filename_lammpstrj())

    @unittest.skipIf(_not_available, _not_available_reason)
    def test_read_frames(self):
        reader = trajectory_reader(self.filename_lammpstrj())
        frames = list(reader)
        self.assertEqual(len(frames), 4)

    @unittest.skipIf(_not_available, _not_available_reason)
    def test_first_frame_contents(self):
        reader = trajectory_reader(self.filename_lammpstrj())
        frame = next(reader)
        self.assertEqual(frame['N'], 24)
        self.assert_arrays_equal_within_float32eps(
            frame['x'][:, 0], self.LAMMPSTRJ_FIRST_FRAME_FIRST_X)

    @unittest.skipIf(_not_available, _not_available_reason)
    def test_first_frame_contents_no_velocities(self):
        reader = trajectory_reader(self.filename_lammpstrj_no_velocities())
        frame = next(reader)
        self.assertEqual(frame['N'], 24)
        self.assertEqual(frame['v'], None)
        self.assert_arrays_equal_within_float32eps(
            frame['x'][:, 0], self.LAMMPSTRJ_FIRST_FRAME_FIRST_X)



__all__ = ['reciprocal_isotropic', 'reciprocal_line']


import logging
import numpy as np

from ctypes import cdll, c_int
from distutils.sysconfig import get_config_var
from numpy import linalg, array, arange, require, nonzero, pi, sqrt, prod
from os.path import dirname, join

logger = logging.getLogger('dynasor')

np_f = dict(d=np.float64, s=np.float32)
np_c = dict(d=np.complex128, s=np.complex64)
np_ndp = np.ctypeslib.ndpointer

_lib = {}
_rho_k = {}
_rho_j_k = {}

ndp_f_2d_r = {}
ndp_c_1d_rw = {}
ndp_c_2d_rw = {}
for t in 'ds':
    ndp_f_2d_r[t] = np_ndp(dtype=np_f[t], ndim=2,
                           flags='f_contiguous, aligned')
    ndp_c_1d_rw[t] = np_ndp(dtype=np_c[t], ndim=1,
                            flags='f_contiguous, aligned, writeable')
    ndp_c_2d_rw[t] = np_ndp(dtype=np_c[t], ndim=2,
                            flags='f_contiguous, aligned, writeable')
    _lib[t] = cdll.LoadLibrary(
        join(dirname(__file__),
             '_rho_j_k_{}{}'.format(t, get_config_var('EXT_SUFFIX'))))
    _lib[t].rho_k.argtypes = (ndp_f_2d_r[t], c_int,
                              ndp_f_2d_r[t], c_int,
                              ndp_c_1d_rw[t])
    _lib[t].rho_j_k.argtypes = (ndp_f_2d_r[t], ndp_f_2d_r[t], c_int,
                                ndp_f_2d_r[t], c_int,
                                ndp_c_1d_rw[t], ndp_c_2d_rw[t])


def calc_rho_k(x, k, ftype='d'):
    """Calculate rho(k) of particle coordinates x.

    Will call external function rho_k to calculate the
    particle density in k-space.
    Particle coordinates and k-space points of interest are
    passed as input via x and k, respectively.
    """
    x = require(x, np_f[ftype], ['F_CONTIGUOUS', 'ALIGNED'])
    k = require(k, np_f[ftype], ['F_CONTIGUOUS', 'ALIGNED'])
    _, Nx = x.shape
    _, Nk = k.shape
    rho_k = np.zeros((Nk,), dtype=np_c[ftype], order='F')
    _lib[ftype].rho_k(x, Nx, k, Nk, rho_k)
    return rho_k


def calc_rho_j_k(x, v, k, ftype='d'):
    """As calc_rho_k, but calculate also velocities in k-space
    """
    assert x.shape == v.shape
    x = require(x, np_f[ftype], ['F_CONTIGUOUS', 'ALIGNED'])
    v = require(v, np_f[ftype], ['F_CONTIGUOUS', 'ALIGNED'])
    k = require(k, np_f[ftype], ['F_CONTIGUOUS', 'ALIGNED'])
    _, Nx = x.shape
    _, Nk = k.shape
    rho_k = np.zeros((Nk,), dtype=np_c[ftype], order='F')
    j_k = np.zeros((3, Nk), dtype=np_c[ftype], order='F')
    _lib[ftype].rho_j_k(x, v, Nx, k, Nk, rho_k, j_k)
    return rho_k, j_k


class reciprocal_processor:
    """Used as a trajectory frame processing helper tool.
    """
    def get_frame_process_function(self):
        """Create a function to be used to process each trajectory frame.

        Augment the frame (which is a dictionary) with k-space densities
        for each particle group (e.g. atom type).
        Depending on whether velocity data is available, calculate and
        add also particle currents.
        """
        def fun(frame):
            frame = frame.copy()
            if 'vs' in frame:
                rho_ks, j_ks = zip(*[calc_rho_j_k(x, v, self.k_points,
                                                  ftype=self.ftype)
                                     for x, v in zip(frame['xs'],
                                                     frame['vs'])])
                jz_ks = [np.sum(j * self.k_direct, axis=0) for j in j_ks]
                frame['j_ks'] = j_ks
                frame['jz_ks'] = jz_ks
                frame['jper_ks'] = [j - (jz * self.k_direct)
                                    for j, jz in zip(j_ks, jz_ks)]
                frame['rho_ks'] = rho_ks
            else:
                frame['rho_ks'] = [calc_rho_k(x, self.k_points,
                                              ftype=self.ftype)
                                   for x in frame['xs']]
            return frame
        return fun

    def process_specific_xs(self, xs):
        return [calc_rho_k(x, self.k_points, ftype=self.ftype) for x in xs]


def get_prune_distance(max_points, max_q, q_vol):
    """Return the prune distance for q/k-points in the isotropic case

    max_points corresponds to the wanted number of resulting q/k-points,
    max_q corresponds to the maximum q-value in the resulting q/k-point set,
    vol_q corresponds to the q-space volume for a single q-point.
    If points are selected from the full grid with probability
    min(1, (q_prune/|q|)^2), k-space will on average be sampled with
    an equal number of points per radial unit (for q > q_prune).
    """
    # Use Cardano's formula to find k_prune
    p = -3.0 * max_q ** 2 / 4
    q = 3.0 * max_points * q_vol / pi - max_q ** 3 / 4
    D = (p / 3) ** 3 + (q / 2) ** 2

    u = (-q / 2 + sqrt(D + 0j)) ** (1.0 / 3)
    v = (-q / 2 - sqrt(D + 0j)) ** (1.0 / 3)
    x = -(u + v) / 2 - 1j * (u - v) * sqrt(3) / 2
    return np.real(x) + max_q / 2


class reciprocal_isotropic(reciprocal_processor):

    def __init__(self, box, max_points=10000, max_k=10.0, ftype='d'):
        """Creates a set of reciprocal coordinates suitable for isotropic
        sampling of k-space. Provide a method to calculate rho_k/j_k
        for trajectory frames.


        Optionally limit the set to approximately max_points points by
        randomly removing points from a "fully populated grid".
        The points are removed in such a way that for k > k_prune,
        the points will be radially uniformely distributed
        (the value of k_prune is calculated from max_k, max_points,
        and the shape of the box).

        Variables named k-something (such as input argument max_k) are expected
        to be "physicist reciprocal length" (i.e. _with_ 2*pi factor).
        Variables named q-something are expected to be without the 2*pi factor.

        ftype can be either 'd' or 's' (double or single precission)
        """

        assert(max_points > 1000)

        self.max_k = max_k
        self.max_points = max_points

        self.ftype = ftype
        npftype = np_f[ftype]
        self.A = require(box.copy(), npftype)
        # B is the "crystallographic" reciprocal vectors
        self.B = linalg.inv(self.A.transpose())

        max_q = max_k / (2.0 * pi)
        q_mins = array([linalg.norm(b) for b in self.B])
        q_vol = prod(q_mins)
        self.q_mins = q_mins

        if max_points > pi * max_q ** 3 / (6 * q_vol):
            # Use all k-points, do not throw any away
            self.q_prune = None
        else:
            self.q_prune = get_prune_distance(max_points, max_q, q_vol)

        b1, b2, b3 = [(2 * pi) * x.reshape((3, 1, 1, 1)) for x in self.B]
        N_k1, N_k2, N_k3 = [int(np.ceil(max_q / dq)) for dq in q_mins]
        k_points = \
            b1 * arange(N_k1, dtype=npftype).reshape((1, N_k1, 1, 1)) + \
            b2 * arange(N_k2, dtype=npftype).reshape((1, 1, N_k2, 1)) + \
            b3 * arange(N_k3, dtype=npftype).reshape((1, 1, 1, N_k3))
        k_points = k_points.reshape((3, int(k_points.size / 3)))
        q_distance = sqrt(np.sum(k_points ** 2, axis=0)) * (1.0 / (2 * pi))

        I, = nonzero(q_distance <= max_q)
        I = I[q_distance[I].argsort()]
        q_distance = q_distance[I]
        k_points = k_points[:, I]  # All k_points < max_k, sorted by length

        if self.q_prune is not None:
            N = len(q_distance)

            # Keep point with probability min(1, (q_prune/|q|)^2) ->
            # aim for an equal number of points per equally thick "onion peel"
            # to get equal number of points per radial unit.
            p = np.ones(N)
            p[1:] = (self.q_prune / q_distance[1:]) ** 2

            I, = nonzero(p > np.random.rand(N))
            q_distance = q_distance[I]
            k_points = k_points[:, I]

        self.k_points = k_points
        self.q_distance = q_distance
        self.k_distance = 2.0 * pi * q_distance
        N = len(q_distance)
        self.k_direct = k_points.copy()
        self.k_direct[:, 1:] /= self.k_distance[1:].reshape((1, N - 1))


class reciprocal_line(reciprocal_processor):
    def __init__(self, points=1000, k_direction=(1.0, 1.0, 1.0), ftype='d'):
        self.ftype = ftype
        npftype = np_f[ftype]
        k_direction = list(k_direction)
        k_direction = require(k_direction, npftype).reshape((3, 1))
        self.max_k = sqrt(np.sum(k_direction ** 2))
        self.k_points = k_direction * np.linspace(0.0, 1.0, points)
        self.k_distance = sqrt(np.sum(self.k_points ** 2, axis=0))
        self.q_distance = self.k_distance * (1.0 / (2 * pi))
        self.k_direct = self.k_points.copy()
        self.k_direct[:, 1:] /= self.k_distance[1:].reshape((1, points - 1))



"""Molfile plugin interface

Not much more than the necessary ctypes stuff.
"""

import os
import re
from sys import maxsize
from platform import uname
from itertools import islice

from ctypes import (
    cdll, CDLL, RTLD_GLOBAL, POINTER, CFUNCTYPE,
    Structure, cast, pointer, c_int, c_uint, c_float, c_double,
    c_char, c_char_p, c_void_p)
from ctypes.util import find_library

_system = uname()[0]
_64bit_python = maxsize > 2 ** 32
if _system != 'Windows':
    # a bit ugly. Is there a more kosher way of making sure that the
    # libstdc++ (or similar) symbols are available to the molfile-plugins?
    _cxx = CDLL(find_library('stdc++'), mode=RTLD_GLOBAL)


c_int_p = POINTER(c_int)
c_int_pp = POINTER(c_int_p)
c_char_pp = POINTER(c_char_p)
c_char_ppp = POINTER(c_char_pp)
c_float_p = POINTER(c_float)
c_float_pp = POINTER(c_float_p)

#
# As of molfile_plugin abiversion 16, the following trajectory
# formats are supported (pasted from the AMD web page):
#
# Molecular Dynamics Trajectory File Plugins
#
#     AMBER 'binpos' trajectory reader (.binpos)
#     AMBER 'CRD' trajectory reader (.crd, .crdbox)
#     CHARMM, NAMD, X-PLOR 'DCD' reader/writer (.dcd)
#     CPMD (CPMD trajectory) reader (.cpmd)
#     DLPOLY HISTORY file reader (.dlpolyhist)
#     Gromacs TRR/XTC reader (.trr, .xtc)
#     LAMMPS trajectory reader (.lammpstrj)
#     MMTK NetCDF trajectory reader (.nc)
#     VASP trajectories of ionic steps (.xml, .OUTCAR, .XCATCAR)
#     VTF trajectory files (.vtf)
#     XCrySDen, Quantum Espresso XSF/AXSF trajectory files (.axsf, .xsf)
#     XYZ trajectory files (.xyz)
#

TRAJECTORY_PLUGIN_MAPPING = (
    # (SOFTWARE-NAME, FILE-TYPE, FILE-SUFFIX, PLUGIN-NAME)
    ('AMBER', '"binpos"',
     'binpos', 'binposplugin'),

    ('AMBER', '"CRD"',
     'crd', 'crdplugin'),

    ('CHARMM', '"DCD" - CHARMM, NAMD, XPLOR',
     'dcd', 'dcdplugin'),

    ('CPMD', 'CPMD',
     'cpmd', 'cpmdplugin'),

    ('DLPOLY', 'DLPOLY History',
     'dlpolyhist', 'dlpolyplugin'),

    ('GROMACS', 'Gromacs XTC',
     'xtc', 'gromacsplugin'),

    ('GROMACS', 'Gromacs TRR',
     'trr', 'gromacsplugin'),

    ('LAMMPS', 'LAMMPS Trajectory',
     'lammpstrj', 'lammpsplugin'),

    ('VASP', 'VASP ionic steps',
     'xml', 'vaspxmlplugin'),

    ('VASP', 'VASP ionic steps',
     'OUTCAR', 'vaspoutcarplugin'),

    ('VTF', 'VTF trajectory',
     'vtf', 'vtfplugin'),

    ('XCrySDen', 'XSF trajectory',
     'xsf', 'xsfplugin'),

    ('XCrySDen', 'AXSF trajectory',
     'axsf', 'xsfplugin'),

    ('?', 'XYZ trajectory',
     'xyz', 'xyzplugin'))


MIN_ABI_VERSION = 15

MOLFILE_PLUGIN_TYPE = 'mol file reader'
VMDPLUGIN_SUCCESS = 0
VMDPLUGIN_ERROR = -1


class vmdplugin_t(Structure):
    _fields_ = [('abiversion', c_int),
                ('type', c_char_p),
                ('name', c_char_p),
                ('prettyname', c_char_p),
                ('author', c_char_p),
                ('majorv', c_int),
                ('minorv', c_int),
                ('is_reentrant', c_int)]


class molfile_metadata_t(Structure):
    _fields_ = [('database', c_char * 81),
                ('accession', c_char * 81),
                ('date', c_char * 81),
                ('title', c_char * 81),
                ('remarklen', c_int),
                ('remarks', c_char_p)]


class molfile_atom_t(Structure):
    _fields_ = [('name', c_char * 16),
                ('type', c_char * 16),
                ('resname', c_char * 16),
                ('resid', c_int),
                ('segid', c_char * 16),
                ('chain', c_char * 16),
                ('altloc', c_char * 16),
                ('insertion', c_char * 16),
                ('occupancy', c_float),
                ('bfactor', c_float),
                ('mass', c_float),
                ('charge', c_float),
                ('radius', c_float),
                ('atomicnumber', c_int)]


class molfile_timestep_metadata_t(Structure):
    _fields_ = [('count', c_uint),
                ('avg_bytes_per_timestamp', c_uint),
                ('has_velocities', c_int)]


class molfile_qm_metadata_t(Structure):
    # Left as an exercise to the reader
    pass


class molfile_qm_timestep_t(Structure):
    # Left as an exercise to the reader
    pass


class molfile_timestep_t(Structure):
    _fields_ = [('coords', POINTER(c_float)),
                ('velocities', POINTER(c_float)),
                ('A', c_float),
                ('B', c_float),
                ('C', c_float),
                ('alpha', c_float),
                ('beta', c_float),
                ('gamma', c_float),
                ('physical_time', c_double)]


class molfile_volumetric_t(Structure):
    _fields_ = [('dataname', c_char * 256),
                ('origin', c_float * 3),
                ('xaxis', c_float * 3),
                ('yaxis', c_float * 3),
                ('zaxis', c_float * 3),
                ('xsize', c_int),
                ('ysize', c_int),
                ('zsize', c_int),
                ('has_color', c_int)]


dummy_fun_t = CFUNCTYPE(c_int)


class molfile_plugin_t(Structure):
    # ABI from molfile abiversion 16
    # This is the important(TM) structure.
    # Only partial read support is considered for now, other
    # functions have been replaced by a dummy_fun_t placeholder.
    _fields_ = [
        ('abiversion', c_int),
        ('type', c_char_p),
        ('name', c_char_p),
        ('prettyname', c_char_p),
        ('author', c_char_p),
        ('majorv', c_int),
        ('minorv', c_int),
        ('is_reentrant', c_int),
        ('filename_extension', c_char_p),
        # void *(* open_file_read)(const char *filepath,
        #                          const char *filetype, int *natoms);
        ('open_file_read', CFUNCTYPE(c_void_p, c_char_p, c_char_p, c_int_p)),
        # int (*read_structure)(void *, int *optflags, molfile_atom_t *atoms);
        ('read_structure', CFUNCTYPE(c_int, c_void_p, c_int_p,
                                     POINTER(molfile_atom_t))),
        # int (*read_bonds)(void *, int *nbonds, int **from, int **to,
        #                   float **bondorder, int **bondtype, int *nbondtypes,
        #                   char ***bondtypename);
        ('read_bonds', dummy_fun_t),
        # int (* read_next_timestep)(void *, int natoms, molfile_timestep_t *);
        ('read_next_timestep', CFUNCTYPE(c_int, c_void_p, c_int,
                                         POINTER(molfile_timestep_t))),
        # void (* close_file_read)(void *);
        ('close_file_read', CFUNCTYPE(None, c_void_p)),
        # void *(* open_file_write)(const char *filepath, const char *filetype,
        #      int natoms);
        ('open_file_write', dummy_fun_t),
        # int (* write_structure)(void *, int optflags,
        #                         const molfile_atom_t *atoms);
        ('write_structure', dummy_fun_t),
        # int (* write_timestep)(void *, const molfile_timestep_t *);
        ('write_timestep', dummy_fun_t),
        # void (* close_file_write)(void *);
        ('close_file_write', dummy_fun_t),
        # int (* read_volumetric_metadata)(void *, int *nsets,
        #                                  molfile_volumetric_t **metadata);
        ('read_volumetric_metadata', CFUNCTYPE(c_int, c_void_p, c_int_p,
                                               POINTER(POINTER(
                                                   molfile_volumetric_t)))),
        # int (* read_volumetric_data)(void *, int set, float *datablock,
        #       float *colorblock);
        ('read_volumetric_data', CFUNCTYPE(c_int, c_void_p, c_int, c_float_p,
                                           c_float_p)),
        # int (* read_rawgraphics)(void *, int *nelem,
        #                          const molfile_graphics_t **data);
        ('read_rawgraphics', dummy_fun_t),
        # int (* read_molecule_metadata)(void *,
        #                                molfile_metadata_t **metadata);
        ('read_molecule_metadata', CFUNCTYPE(c_int, c_void_p,
                                             POINTER(POINTER(
                                                molfile_metadata_t)))),
        # int (* write_bonds)(void *, int nbonds, int *from, int *to,
        #                    float *bondorder, int *bondtype, int nbondtypes,
        #                    char **bondtypename);
        ('write_bonds', dummy_fun_t),
        # int (* write_volumetric_data)(void *,
        #                               molfile_volumetric_t *metadata,
        #                               float *datablock, float *colorblock);
        ('write_volumetric_data', dummy_fun_t),
        # int (* read_angles)(void *handle, int *numangles, int **angles,
        #                     int **angletypes, nt *numangletypes,
        #                     char ***angletypenames, int *numdihedrals,
        #                     int **dihedrals, int **dihedraltypes,
        #                     int *numdihedraltypes, char ***dihedraltypenames,
        #                     int *numimpropers, int **impropers,
        #                     int **impropertypes, int *numimpropertypes,
        #                     char ***impropertypenames, int *numcterms,
        #                     int **cterms, int *ctermcols, int *ctermrows)
        ('read_angles', dummy_fun_t),
        # int (* write_angles)(void *handle, int numangles, const int *angles,
        #                      const int *angletypes, int numangletypes,
        #                      const char **angletypenames, int numdihedrals,
        #                      const int *dihedrals, const int *dihedraltypes,
        #                      int numdihedraltypes,
        #                      const char **dihedraltypenames,
        #                      int numimpropers, const int *impropers,
        #                      const int *impropertypes, int numimpropertypes,
        #                      const char **impropertypenames, int numcterms,
        #                      const int *cterms, int ctermcols, int ctermrows)
        ('write_angles', dummy_fun_t),
        # int (* read_qm_metadata)(void *, molfile_qm_metadata_t *metadata);
        ('read_qm_metadata', dummy_fun_t),
        # int (* read_qm_rundata)(void *, molfile_qm_t *qmdata);
        ('read_qm_rundata', dummy_fun_t),
        # int (* read_timestep)(void *, int natoms, molfile_timestep_t *,
        #                       molfile_qm_metadata_t *,
        #                       molfile_qm_timestep_t *);
        ('read_timestep', CFUNCTYPE(c_int, c_void_p, c_int,
                                    POINTER(molfile_timestep_t),
                                    POINTER(molfile_qm_metadata_t),
                                    POINTER(molfile_qm_timestep_t))),
        # int (* read_timestep_metadata)(void *, molfile_timestep_metadata_t *)
        ('read_timestep_metadata',
            CFUNCTYPE(c_int, c_void_p, POINTER(molfile_timestep_metadata_t))),
        # int (* read_qm_timestep_metadata)(void *,
        #                                   molfile_qm_timestep_metadata_t *);
        ('read_qm_timestep_metadata', dummy_fun_t),
        #  int (* cons_fputs)(const int, const char*);
        ('cons_fputs', CFUNCTYPE(c_int, c_int, c_char_p))]


# typedef int (*vmdplugin_register_cb)(void *, vmdplugin_t *);
vmdplugin_register_cb_t = CFUNCTYPE(c_int, c_void_p, POINTER(vmdplugin_t))


class _MolfilePlugin(object):
    """A thin molfile_plugin wrapper class

    This class holds the loaded plugin-library and sets up
    a molfile_plugin_t structure.

    The 'close' method, calls the plugin fini-function.
    """
    def __init__(self, plugin_path):
        library = self._load_plugin_library(plugin_path)

        if library.vmdplugin_init() != VMDPLUGIN_SUCCESS:
            raise RuntimeError('Failed to init %s' % plugin_path)

        plugin_p = pointer(molfile_plugin_t(0))
        callback = self._create_vmdplugin_register_cb(plugin_p)

        if library.vmdplugin_register(None, callback) != VMDPLUGIN_SUCCESS:
            raise RuntimeError('Failed to register %s' % plugin_path)

        self._library = library
        self.plugin = plugin_p.contents

    def close(self):
        self.plugin = molfile_plugin_t(0)
        self._library.vmdplugin_fini()

    def _load_plugin_library(self, plugin_path):
        library = cdll.LoadLibrary(plugin_path)

        # extern int vmdplugin_init(void);
        library.vmdplugin_init.restype = c_int

        # extern int vmdplugin_fini(void);
        library.vmdplugin_fini.restype = c_int

        # extern int vmdplugin_register(void *, vmdplugin_register_cb);
        library.vmdplugin_register.restype = c_int
        library.vmdplugin_register.argtypes \
            = (c_void_p, vmdplugin_register_cb_t)

        return library

    def _create_vmdplugin_register_cb(self, plugin_p):
        def py_register_callback(v, p):
            contents = p.contents
            if contents.type == MOLFILE_PLUGIN_TYPE:
                if contents.abiversion >= MIN_ABI_VERSION:
                    plugin_p.contents \
                        = cast(p, POINTER(molfile_plugin_t)).contents
            return VMDPLUGIN_SUCCESS

        vmdplugin_register_cb = vmdplugin_register_cb_t(py_register_callback)

        return vmdplugin_register_cb


def molfile_plugin_path(plugin_name):
    plugin_dir = molfile_plugin_dir()
    plugin_suffix = _molfile_machine_specific_plugin_suffix()
    path = os.path.join(plugin_dir, plugin_name + '.' + plugin_suffix)
    if os.path.exists(path):
        return path


def molfile_plugin_dir():
    vmddir = _vmddir()
    if vmddir is not None:
        return os.path.join(vmddir, _molfile_relative_plugin_dir())


def _molfile_relative_plugin_dir():
    'Molfile path relative VMD root'
    return os.path.join('plugins',
                        _molfile_machine_specific_plugin_dir_name(),
                        'molfile')


def _molfile_machine_specific_plugin_suffix():
    return 'so'


def _molfile_machine_specific_plugin_dir_name():
    if _system == 'Linux':
        if _64bit_python:
            return 'LINUXAMD64'
        else:
            return 'LINUX'

    elif not _64bit_python:
        if _system == 'Darwin':
            return 'MACOSXX86'

        elif _system == 'Windows':
            return 'WIN32'

    bits = '64' if _64bit_python else '32'
    raise RuntimeError('No known {}-bit plugin-path on {} system'
                       .format(bits, _system))


def _vmddir():
    vmddir = os.environ.get('VMDDIR')
    if vmddir is not None:
        return vmddir

    for path in os.environ['PATH'].split(os.pathsep):
        tentative_vmd_script_path = os.path.join(path, 'vmd')
        if os.path.exists(tentative_vmd_script_path):
            vmddir = _get_vmddir_from_vmd_script(tentative_vmd_script_path)
            if vmddir:
                return vmddir

    for tentive_vmddir in ['C:\\Program Files (x86)\\University of Illinois\\VMD']:  # noqa
        if os.path.exists(tentive_vmddir):
            return tentive_vmddir


def _get_vmddir_from_vmd_script(vmd_script_path):
    with open(vmd_script_path) as fh:
        for L in islice(fh, 10):
            match = re.match(r'^(?:set )?defaultvmddir=(?:"(/.*)"|(/[^# ]*)).*$', L)  # noqa
            if match:
                a, b = match.groups()
                return a or b



import numpy
import os

class TrajectoryReaderTestMixin(object):

    LAMMPSTRJ_FIRST_FRAME_FIRST_X = numpy.array([0.191468, 0.302071, 0.0528818])

    def filename_lammpstrj(self):
        data_path = self._data_dir_path()
        return os.path.join(data_path, "positions_and_velocities.lammpstrj")

    def filename_lammpstrj_no_velocities(self):
        data_path = self._data_dir_path()
        return os.path.join(data_path, "positions.lammpstrj")

    def filename_xtc_1frame_3atoms(self):
        data_path = self._data_dir_path()
        return os.path.join(data_path), "1frame3atoms.xtc"
    
    def assert_arrays_equal_within_float32eps(self, a, b):
        abs_diff = numpy.absolute(a - b)
        eps = self._float32abs()
        self.assertTrue((abs_diff < eps).all(),
                        "%s and %s not within eps from each other" % (a, b))

    def _data_dir_path(self):
        this_dir = os.path.dirname(__file__)
        return os.path.join(this_dir, "data")

    def _float32abs(self):
        return numpy.finfo(numpy.float32).eps



import unittest
from dsf.trajectory_reader.xtc_trajectory_reader import (
    xtc_trajectory_reader as trajectory_reader)
from dsf.trajectory_reader.test.trajectory_reader_test_mixin import TrajectoryReaderTestMixin


_not_available = not trajectory_reader.available()
_not_available_reason = "gromacs xtc plugin not available"

class XTCTrajectoryReaderTest(unittest.TestCase, TrajectoryReaderTestMixin):

    @unittest.skipIf(_not_available, _not_available_reason)
    def test_open_xtc(self):
        trajectory_reader(self.filename_xtc_1frame_3atoms())

    @unittest.skipIf(_not_available, _not_available_reason)
    def test_read_frames(self):
        reader = trajectory_reader(self.filename_xtc_1frame_3atoms())
        frames = list(reader)
        self.assertEqual(len(frames), 1)

    @unittest.skipIf(_not_available, _not_available_reason)
    def test_first_frame_contents(self):
        reader = trajectory_reader(self.filename_xtc_1frame_3atoms())
        frame = reader.next()
        self.assertEqual(frame['N'], 3)
        self.assertEqual(frame['v'], None)
        self.assert_arrays_equal_within_float32eps(frame['x'][:, 0],
                                                   self.XTC_FIRST_FRAME_FIRST_X)



import numpy as np
from dsf.trajectory_reader.abstract_trajectory_reader \
    import abstract_trajectory_reader
from dsf.trajectory_reader.molfile_plugin \
    import (_MolfilePlugin, molfile_plugin_dir, molfile_plugin_path,
            TRAJECTORY_PLUGIN_MAPPING, molfile_atom_t,
            molfile_timestep_metadata_t, molfile_timestep_t)
from ctypes import c_float, c_int, byref, POINTER
from itertools import count
from numpy import zeros, pi, array, cos


# Molfile plugin is using single precission floats
molfile_float_np = np.float32
molfile_float_ct = c_float


class molfile_trajectory_reader(abstract_trajectory_reader):
    """Read a trajectory using the molfile_plugin package

    molfile_plugin is a part of VMD, and consists of plugins for a fairly large
    number of different trajectory formats (see
    molfile_plugin.TRAJECTORY_PLUGIN_MAPPING).

    filename - string, filename of trajectory file. index_file - string,
    filename of ini-style index file. plugin - string, name of plugin to use.
    If None, guess  pluginname by looking at filename suffix. """

    @classmethod
    def available(cls):
        return molfile_plugin_dir() is not None

    def _guess_plugin(self, filename):
        filename_suffix = filename.rsplit('.', 1)[-1]
        for _, _, suffix, plugin_name in TRAJECTORY_PLUGIN_MAPPING:
            if suffix == filename_suffix:
                return plugin_name

    def __init__(self, filename, plugin_name=None, x_factor=0.1, t_factor=1.0):

        if plugin_name is None:
            plugin_name = self._guess_plugin(filename)

        if plugin_name is None:
            raise RuntimeError('molfile_reader: no suitable plugin known for'
                               ' file {}'.format(filename))

        self.x_factor = x_factor
        self.t_factor = t_factor
        self.v_factor = x_factor / t_factor

        self._N = c_int()
        suffix = filename.rsplit('.', 1)[-1]

        self._mfp = _MolfilePlugin(molfile_plugin_path(plugin_name))
        p = self._mfp.plugin

        self._fh = p.open_file_read(filename, suffix,
                                    byref(self._N))
        if not self._fh:
            raise RuntimeError('molfile_reader: failed to open file {} with'
                               ' plugin {}.'.format(filename, plugin_name))
        N = self._N.value

        if p.read_structure:
            # for e.g. lammpsplugin, read_structure needs to be called first so
            # that molfile_reader knows (internally) which coordinates and
            # velocites to map to which atoms.
            self._atoms_arr = (molfile_atom_t * N)()
            self._optflags = c_int()
            rc = p.read_structure(self._fh, byref(self._optflags),
                                  self._atoms_arr)
            if rc:
                raise IOError('molfile_reader: read structure failed for'
                              ' file {} (plugin {}, rc {})'
                              .format(filename, plugin_name, rc))
        else:
            self._atoms_arr = None

        self._v = None
        self._x = np.require(zeros((3, N)), molfile_float_np,
                             ['F_CONTIGUOUS', 'ALIGNED'])
        self._x.flags.writeable = False

        if p.read_timestep_metadata:
            # It seems only lammpsplugin offers this (but other formats could
            # include velocity information. how to test for that!?)
            tsm = self._timestep_metadata = molfile_timestep_metadata_t()
            rc = p.read_timestep_metadata(self._fh, byref(tsm))
            if rc:
                raise IOError('molfile_reader: read timestep metadata failed'
                              ' for file {} (plugin {}, rc {})'
                              .format(filename, plugin_name, rc))

            if tsm.has_velocities:
                self._v = np.require(zeros((3, N)), molfile_float_np,
                                     ['F_CONTIGUOUS', 'ALIGNED'])
                self._v.flags.writeable = False
        else:
            self._timestep_metadata = None

        # Now, set up the timestep structure
        self._ts = molfile_timestep_t()
        self._ts.coords = self._x.ctypes.data_as(POINTER(molfile_float_ct))
        if self._v is not None:
            self._ts.velocities \
                = self._v.ctypes.data_as(POINTER(molfile_float_ct))
        else:
            # Set velocities to a NULL pointer
            self._ts.velocities = POINTER(molfile_float_ct)()

        # Set frame counter
        self._index = count(1)

    def __iter__(self):
        return self

    def __next__(self):
        if not self._mfp.plugin.read_next_timestep:
            raise StopIteration

        ts = self._ts
        rc = self._mfp.plugin.read_next_timestep(self._fh, self._N, byref(ts))
        if rc:
            self._mfp.close()
            raise StopIteration

        res = dict(index=next(self._index),
                   box=self._to_box(ts.A, ts.B, ts.C, ts.alpha, ts.beta,
                                    ts.gamma) * self.x_factor,
                   N=self._N.value,
                   time=ts.physical_time * self.t_factor,
                   x=self._x * self.x_factor
                   )
        if self._v is not None:
            res['v'] = self._v * self.v_factor
        else:
            res['v'] = None

        return res

    def close(self):
        self._mfp.close()

    def _to_box(self, A, B, C, alpha, beta, gamma):
        # create box vectors out of molfile-info
        deg2rad = pi / 180.0
        return array(((A, 0.0, 0.0),
                      (B * cos(deg2rad * gamma), B, 0.0),
                      (C * cos(deg2rad * beta), C * cos(deg2rad * alpha), C)))



from ctypes import cdll, POINTER, c_float, c_int, c_char_p
from ctypes.util import find_library
from itertools import count
from dsf.trajectory_reader.abstract_trajectory_reader \
    import abstract_trajectory_reader
import numpy as np

#
# L I B G M X
#
# libgmx comes with Gromacs and contains, among other things,
# functionality for reading xtc-files.
#

libgmx_name = find_library('gromacs')
libgmx = libgmx_name and cdll.LoadLibrary(libgmx_name)
np_ndp = np.ctypeslib.ndpointer

if libgmx:
    # single prec gmx-real equals float, right?
    xtcfloat_np = np.float32
    xtcfloat_ct = c_float
    xtcint_ct = c_int

    # t_fileio *open_xtc(const char *filename,const char *mode);
    # /* Open a file for xdr I/O */
    libgmx.open_xtc.restype = POINTER(xtcint_ct)
    libgmx.open_xtc.argtypes = [c_char_p, c_char_p]

    # int read_first_xtc(t_fileio *fio,
    #                           int *natoms,int *step,real *time,
    #                           matrix box,rvec **x,real *prec,gmx_bool *bOK);
    # /* Open xtc file, read xtc file first time, allocate memory for x */
    libgmx.read_first_xtc.restype = xtcint_ct
    libgmx.read_first_xtc.argtypes = [
        POINTER(xtcint_ct), POINTER(xtcint_ct),
        POINTER(xtcint_ct), POINTER(xtcfloat_ct),
        np_ndp(dtype=xtcfloat_np, shape=(3, 3),
               flags='f_contiguous, aligned'),
        POINTER(POINTER(xtcfloat_ct)),
        POINTER(xtcfloat_ct), POINTER(xtcint_ct)]

    # int read_next_xtc(t_fileio *fio,
    #                          int natoms,int *step,real *time,
    #                          matrix box,rvec *x,real *prec,gmx_bool *bOK);
    # /* Read subsequent frames */
    libgmx.read_next_xtc.restype = xtcint_ct
    libgmx.read_next_xtc.argtypes = [
        POINTER(xtcint_ct), xtcint_ct,
        POINTER(xtcint_ct), POINTER(xtcfloat_ct),
        np_ndp(dtype=xtcfloat_np, shape=(3, 3),
               flags='f_contiguous, aligned'),
        np_ndp(dtype=xtcfloat_np, ndim=2,
               flags='f_contiguous, aligned'),
        POINTER(xtcfloat_ct), POINTER(xtcint_ct)]


class xtc_trajectory_reader(abstract_trajectory_reader):

    @classmethod
    def available(cls):
        # TODO: Fix so that this works also for gromacs 5
        # When changing from libgmx to libgromacs-5,
        # it seems the library function signatures now
        # include an XDR-data structore.
        # Investigate, implement, and test...
        return False
#         return libgmx is not None

    def __init__(self, filename):
        if libgmx is None:
            raise RuntimeError('XTC_reader: No libgmx found, cannot use XTC_reader!')

        self._fio = libgmx.open_xtc(filename, 'rb')
        if not self._fio:
            raise IOError('XTC_reader: Failed to open file %s' % filename)

        self._index = count(1)
        self._natoms = xtcint_ct()
        self._step = xtcint_ct()
        self._time = xtcfloat_ct()
        self._box = np.require(np.zeros((3, 3)),
                               xtcfloat_np, ['F_CONTIGUOUS', 'ALIGNED'])
        self._x = None
        self._prec = xtcfloat_ct()
        self._bOK = xtcint_ct()  # gmx_bool equals int
        self._open = True
        self._first_called = False

    def _get_first(self):
        # Read first frame and update state of self accordingly
        _xfirst = POINTER(xtcfloat_ct)()
        res = libgmx.read_first_xtc(self._fio, self._natoms,
                                    self._step, self._time,
                                    self._box, _xfirst,
                                    self._prec, self._bOK)
        self._first_called = True
        if not res:
            raise IOError('XTC_reader: read_first_xtc failed')
        if not self._bOK.value:
            raise IOError('XTC_reader: corrupt frame in xtc-file?')

        N = self._natoms.value
        self._x = np.require(
            np.array(_xfirst[0:3 * N]).reshape((3, N), order='F'),
            xtcfloat_np, ['F_CONTIGUOUS', 'ALIGNED'])
        self._x.flags.writeable = False

    def _get_next(self):
        # get next frame, update state of self
        res = libgmx.read_next_xtc(self._fio, self._natoms.value,
                                   self._step, self._time,
                                   self._box, self._x,
                                   self._prec, self._bOK)
        if not res:
            return False
        if not self._bOK.value:
            raise IOError('XTC_reader: corrupt frame in xtc-file?')
        return True

    def __iter__(self):
        return self

    def close(self):
        if self._open:
            libgmx.close_xtc(self._fio)
            self._open = False

    def __next__(self):
        if not self._open:
            raise StopIteration

        if self._first_called:
            if not self._get_next():
                self.close()
                raise StopIteration
        else:
            self._get_first()

        return dict(
            index=next(self._index),
            box=self._box.copy('F'),
            time=self._time.value,
            N=self._natoms.value,
            x=self._x,
            v=None,
            )



"""
This module provides an implementation of Filon's integration formula.
For information about Filon's formula, see e.g.
`Abramowitz and Stegun, Handbook of Mathematical Functions,
section 25 <http://mathworld.wolfram.com/FilonsIntegrationFormula.html>`_ or
Allen and Tildesley, *Computer Simulation of Liquids*, Appendix D.

Integration is performed along one dimension (default ``axis=0``), e.g.,

.. code::

    [F0[0]  F1[0] ..  FN[0] ]     [f0[0]  f1[0] ..  fN[0] ]
    [   .      .         .  ]     [   .      .         .  ]
    [F0[.]  F1[.] ..  FN[.] ] = I([f0[.]  f1[.] ..  fN[.] ], dx, [k[0] .. k[Nk]])
    [   .      .         .  ]     [   .      .         .  ]
    [F0[Nk] F1[Nk] .. FN[Nk]]     [f0[Nx] f1[Nx] .. fN[Nx]]

where ``k`` and ``Fj`` have end index ``Nk``, and ``fj`` has end index ``Nx``.
``Nk`` is automatically set by the length of ``k``. Due to the algorithm,
``fj[Nx]`` must be of odd length (``Nx`` must be an even number), and should
correspond to a linearly spaced set of data points (separated by `dx` along the
integration axis).

:func:`sin_integral` and :func:`cos_integral` allow for shifted integration
intervals by the optional argument ``x0``.

"""

__all__ = ['fourier_cos', 'sin_integral', 'cos_integral']

from numpy import (sin, cos, pi, mod, zeros, arange,
                   linspace, require, where, sum)


def fourier_cos(f, dx, k=None, axis=0):
    """Calculates the direct Fourier cosine transform :math:`F(k)` of a
    function :math:`f(x)` using Filon's integration method.

    The array values ``f[0]..f[2n]`` are expected to correspond to
    :math:`f(0.0)\ldots f(2n\Delta x)`. Hence, ``f`` should contain an odd
    number of elements.

    The transform is approximated by the integral
    :math:`F(k) = 2\int_{0}^{x_{max}} f(x) \cos(k x) dx`,
    where :math:`x_{max} = 2n \Delta x`.

    Parameters
    ----------
    f : array
        function values; must contain an odd number of elements
    dx : float
        spacing of x-axis (:math:`\Delta x`)
    k : array
        values of reciprocal axis, at which to evaluate transform;
        if ``k`` is not provided, ``linspace(0.0, 2*pi/dx, f.shape[axis])``,
        will be used.
    axis : int
        axis along which to carry out integration

    Returns
    -------
    (array, array)
        tuple of ``k`` and ``F`` values

    Example
    -------
    A common use case is

    .. code-block:: python

        k, F = fourier_cos(f, dx)
    """

    if k is None:
        k = linspace(0.0, 2 * pi / dx, f.shape[axis])

    return k, 2 * cos_integral(f, dx, k, x0=0.0, axis=axis)


def cos_integral(f, dx, k, x0=0.0, axis=0):
    """Calculates the integral
    :math:`\int_{x_0}^{2n\Delta x} f(x) \cos(k x) dx`.

    Parameters
    ----------
    f : array
        function values; must contain an odd number of elements
    dx : float
        spacing of x-axis (:math:`\Delta x`)
    k : array
        values of reciprocal axis, at which to evaluate transform;
        if ``k`` is not provided, ``linspace(0.0, 2*pi/dx, f.shape[axis])``,
        will be used.
    x0 : float
        offset for integration interval
    axis : int
        axis along which to carry out integration

    Returns
    -------
    float
        tuple of ``k`` and ``F`` values
    """
    return _gen_sc_int(f, dx, k, x0, axis, cos)


def sin_integral(f, dx, k, x0=0.0, axis=0):
    """Calculates the integral
    :math:`\int_{x_0}^{2n\Delta x} f(x) \sin(k x) dx`.

    Parameters
    ----------
    f : array
        function values; must contain an odd number of elements
    dx : float
        spacing of x-axis (:math:`\Delta x`)
    k : array
        values of reciprocal axis, at which to evaluate transform;
        if ``k`` is not provided, ``linspace(0.0, 2*pi/dx, f.shape[axis])``,
        will be used.
    x0 : float
        offset for integration interval
    axis : int
        axis along which to carry out integration

    Returns
    -------
    float
        tuple of ``k`` and ``F`` values
    """
    return _gen_sc_int(f, dx, k, x0, axis, sin)


def _gen_sc_int(f, dx, k, x0, axis, sc):

    f = require(f)
    k = require(k)

    try:
        axis = range(f.ndim)[axis]
    except (IndexError, TypeError):
        raise Exception('axis(={}) is invalid'.format(str(axis)))

    if k.ndim != 1:
        raise ValueError('k must be one-dimensional')

    Nk = len(k)
    Nx = f.shape[axis]
    x_shape = [1] * f.ndim + [Nx]  # We'll transpose axis and put it last
    k_shape = [1] * (f.ndim + 1)  #
    k_shape[axis] = Nk  #

    if mod((Nx - 1), 2) != 0 or Nx < 3:
        raise ValueError('len f must be odd and >2 along its integration axis')

    s = (slice(None),) * f.ndim
    odd_index = s + (slice(1, None, 2),)
    even_index = s + (slice(0, None, 2),)
    first_index = s + ((0,),)
    last_index = s + ((-1,),)

    alpha, beta, gamma = [x.reshape(k_shape[:-1])
                          for x in _alpha_beta_gamma(dx * k)]
    x = (x0 + dx * arange(0.0, Nx)).reshape(x_shape)
    k = k.reshape(k_shape)

    # Add an extra dimension to f, and transpose it to put the x-dimension at
    # axis=-1
    t = list(range(f.ndim + 1))
    t[axis], t[-1] = t[-1], t[axis]
    f = f.reshape(f.shape + (1,)).transpose(t)

    sc_k_x = sc(k * x)
    sc_k_x[first_index] *= 0.5
    sc_k_x[last_index] *= 0.5

    if sc == sin:
        return dx * (alpha * sum((f[first_index] * cos(k * x0) -
                                  f[last_index] * cos(k * x[last_index])),
                                 axis=-1) +
                     beta * sum(f[even_index] * sc_k_x[even_index], axis=-1) +
                     gamma * sum(f[odd_index] * sc_k_x[odd_index], axis=-1))
    elif sc == cos:
        return dx * (alpha * sum((f[last_index] * sin(k * x[last_index]) -
                                  f[first_index] * sin(k * x0)), axis=-1) +
                     beta * sum(f[even_index] * sc_k_x[even_index], axis=-1) +
                     gamma * sum(f[odd_index] * sc_k_x[odd_index], axis=-1))

    raise RuntimeError('Internal error, this should not happen')


def _alpha_beta_gamma(theta):
    # From theta, calculate alpha, beta, and gamma

    N = len(theta)
    alpha = zeros(N)
    beta = zeros(N)
    gamma = zeros(N)

    # theta==0 needs special treatment
    I_nz, = theta.nonzero()
    I_z, = where(theta == 0.0)
    if len(I_z) > 0:
        beta[I_z] = 2.0 / 3.0
        gamma[I_z] = 4.0 / 3.0
        theta = theta[I_nz]

    sin_t = sin(theta)
    cos_t = cos(theta)
    sin2_t = sin_t * sin_t
    cos2_t = cos_t * cos_t
    theta2 = theta * theta
    itheta3 = 1.0 / (theta2 * theta)

    alpha[I_nz] = itheta3 * (theta2 + theta * sin_t * cos_t - 2 * sin2_t)
    beta[I_nz] = 2 * itheta3 * (theta * (1 + cos2_t) - 2 * sin_t * cos_t)
    gamma[I_nz] = 4 * itheta3 * (sin_t - theta * cos_t)

    return alpha, beta, gamma



__all__ = ['create_mfile', 'create_pfile']

import sys
import numpy as np
import logging

logger = logging.getLogger('dynasor')


def create_mfile(filename, output, comment=None):
    """ Creates matlab style m-file. """
    with open(filename, 'w') as fh:
        popts = np.get_printoptions()
        np.set_printoptions(threshold=sys.maxsize, linewidth=sys.maxsize)

        if comment is not None:
            fh.write('%%%\n')
            fh.write(''.join(['% '+x+'\n' for x in comment.split('\n')]))
            fh.write('%%%\n')

        for v, n, desc in output:
            fh.write("\n%% %s\n%s = ...\n%s;\n" % (desc, n, str(v)))

        np.set_printoptions(**popts)
        logger.info('Wrote Matlab-style output to {}'.format(fh.name))


def create_pfile(filename, output, comment=None):
    """ Creates python pickle file. """
    import pickle
    with open(filename, 'wb') as fh:
        pickle.dump(output, fh)
        logger.info('Wrote pickled output to {}'.format(fh.name))



class trajectory_frame(object):
    """Trivial data struct holding MD-data for one time frame

     'index' : trajectory index,
     'box'   : simulation box as 3 row vectors (nm),
     'N'     : number of atoms,
     'x'     : particle positions as 3xN array (nm),
     'v'     : (*) particle velocities as 3xN array (nm/ps),
     'time'  : (*) simulation time (ps),

     (*) may not be available, depends on reader and trajectory file format.
    """
    def __init__(self, **kwargs):
        for attribute, value in kwargs.iteritems():
            setattr(self, attribute, value)



from abc import abstractmethod, ABCMeta


class abstract_trajectory_reader:
    __metaclass__ = ABCMeta

    """Provides a way to iterate through a molecular dynamics (MD) trajectory
    file.

    Each frame/time-step is returned as a trajectory_frame.
    """

    @classmethod
    @abstractmethod
    def available(cls):
        """ Returns True if this reader available on a particular system. """
        pass

    @abstractmethod
    def __iter__(self):
        """ Iterates through the trajectory file, frame by frame. """
        pass

    @abstractmethod
    def __next__(self):
        """ Gets next trajectory frame. """
        pass

    @abstractmethod
    def close(self):
        """ Closes down, release resources etc. """
        pass



import unittest
from dsf.trajectory_reader.molfile_trajectory_reader import (
    molfile_trajectory_reader as trajectory_reader)
from dsf.trajectory_reader.test.trajectory_reader_test_mixin import TrajectoryReaderTestMixin

_not_available = not trajectory_reader.available()
_not_available_reason = "Molfile plugin not available"


class MolfileTrajectoryReaderTest(unittest.TestCase, TrajectoryReaderTestMixin):

    @unittest.skipIf(_not_available, _not_available_reason)
    def test_open_lammpstrj_no_velocities(self):
        trajectory_reader(self.filename_lammpstrj_no_velocities())

    @unittest.skipIf(_not_available, _not_available_reason)
    def test_open_lammpstrj(self):
        trajectory_reader(self.filename_lammpstrj())

    @unittest.skipIf(_not_available, _not_available_reason)
    def test_read_frames(self):
        reader = trajectory_reader(self.filename_lammpstrj())
        frames = list(reader)
        self.assertEqual(len(frames), 4)

    @unittest.skipIf(_not_available, _not_available_reason)
    def test_first_frame_contents(self):
        reader = trajectory_reader(self.filename_lammpstrj())
        frame = reader.next()
        self.assertEqual(frame['N'], 24)
        self.assert_arrays_equal_within_float32eps(frame['x'][:, 0],
                                                   self.LAMMPSTRJ_FIRST_FRAME_FIRST_X)

    @unittest.skipIf(_not_available, _not_available_reason)
    def test_first_frame_contents_no_velocities(self):
        reader = trajectory_reader(self.filename_lammpstrj_no_velocities())
        frame = reader.next()
        self.assertEqual(frame['N'], 24)
        self.assertEqual(frame['v'], None)
        self.assert_arrays_equal_within_float32eps(frame['x'][:, 0],
                                                   self.LAMMPSTRJ_FIRST_FRAME_FIRST_X)



#!/usr/bin/env python

import re
import sys

from config import (extra_compile_args, extra_link_args,
                    local_compiler, local_link_shared, local_linker)
try:
    from setuptools import setup, Extension
except ImportError:
    print('setuptools not found; falling back to distutils')
    from distutils.core import setup
    from distutils.extension import Extension
from distutils.sysconfig import get_config_vars


if sys.version_info < (3, 6, 0, 'final', 0):
    raise SystemExit('Python 3.6 or later is required!')

with open('README.rst', encoding='utf-8') as fd:
    long_description = fd.read()


if local_compiler is not None:
    # Kludge: Force compiler of choice for building _rho_j_k.c.
    # _rho_j_k.c only contains a plain c-function with no python-
    # dependencies at all. Hence, just blatantly set simple
    # compiler, linker and flags (inspired by GPAW setup.py).

    config_vars = get_config_vars()
    for key in ['BASECFLAGS', 'CFLAGS', 'OPT', 'PY_CFLAGS',
                'CCSHARED', 'CFLAGSFORSHARED', 'LINKFORSHARED',
                'LIBS', 'SHLIBS']:
        config_vars[key] = ''

    config_vars['CC'] = local_compiler
    config_vars['LDSHARED'] = ' '.join([local_linker] + local_link_shared)

rho_j_k_d_ext = Extension('dsf._rho_j_k_d',
                          sources=['src/_rho_j_k.c'],
                          define_macros=[('RHOPREC', 'double')],
                          extra_compile_args=extra_compile_args,
                          extra_link_args=extra_link_args,
                          )

rho_j_k_s_ext = Extension('dsf._rho_j_k_s',
                          sources=['src/_rho_j_k.c'],
                          define_macros=[('RHOPREC', 'float')],
                          extra_compile_args=extra_compile_args,
                          extra_link_args=extra_link_args,
                          )


with open('dsf/__init__.py') as fd:
    lines = '\n'.join(fd.readlines())
version = re.search("__version__ = '(.*)'", lines).group(1)
maintainer = re.search("__maintainer__ = '(.*)'", lines).group(1)
maintainer_email = re.search("__maintainer_email__ = '(.*)'", lines).group(1)
url = re.search("__url__ = '(.*)'", lines).group(1)
license = re.search("__license__ = '(.*)'", lines).group(1)
description = re.search("__description__ = '(.*)'", lines).group(1)

# PyPI name
name = 'dynasor'
# Linux-distributions may want to change the name:
if 0:
    name = 'python-dynasor'

setup(name=name,
      version=version,
      description=description,
      long_description=long_description,
      url=url,
      maintainer=maintainer,
      maintainer_email=maintainer_email,
      packages=['dsf', 'dsf/trajectory_reader'],
      include_package_data=True,
      ext_modules=[rho_j_k_d_ext, rho_j_k_s_ext],
      scripts=['dynasor'],
      install_requires=['numpy'],
      platforms=['unix'],
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3 :: Only',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Intended Audience :: Education',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python',
          'Programming Language :: C',
          'Topic :: Scientific/Engineering :: Chemistry',
          'Topic :: Scientific/Engineering :: Physics'])



__all__ = ['get_itraj', 'iwindow']

import logging
import sys

from collections import deque
from itertools import islice
from os.path import isfile

from .trajectory_reader import available_readers

logger = logging.getLogger('dynasor')


def get_itraj(filename: str, step: int=1, max_frames: int=0,
              readers=available_readers):
    """Return a dynasor-style trajectory iterator

    Simple wrapper for the trajectory_reader-classes.

    Parameters
    ----------
    filename : str
        name of input file
    step : int
        step to access (1 by default = every single frame); must be > 0.
    max_frames : int
        maximum number of frames to read (0 by default = no limit);
        must be >= 0.

    Returns
    -------
    Each iterator step consists of a dictionary.


    .. code-block:: python

        {
         'index' : trajectory frame index (1, 2, 3, ...),
         'box'   : simulation box as 3 row vectors (nm),
         'N'     : number of atoms,
         'x'     : particle positions as 3xN array (nm),
         'v'     : (*) particle velocities as 3xN array (nm/ps),
         'time'  : (*) simulation time (ps),
        }

    ``(*)`` may not be available, depends on reader and trajectory file format.
    """

    assert step > 0
    assert max_frames >= 0
    if max_frames == 0:
        max_frames = sys.maxsize
    elif step > 1:
        max_frames = max_frames * step

    if not isfile(filename):
        raise IOError('File {} does not exist'.format(filename))

    for reader in readers:
        # Simply pick the first reader that seems to work
        if reader.available():
            reader_name = reader.__name__
            try:
                logger.debug('Trying trajectory_reader %s' % reader_name)
                return islice(reader(filename), 0, max_frames, step)
            except Exception:
                logger.debug('Trying trajectory_reader {}'
                             ' failed to open file {}'
                             .format(reader_name, filename))

    raise IOError('Failed to open trajectory file {}'.format(filename))


def consume(iterator, n):
    """ Advance the iterator by n steps. If n is none, consume entirely. """
    # From the python.org
    if n is None:
        deque(iterator, maxlen=0)
    else:
        next(islice(iterator, n, n), None)


class iwindow:
    """Sliding window iterator.

    Returns consecutive windows (a windows is represented as a list
    of objects), created from an input iterator.

    Parameters
    ----------
    width : int
        length of window
    stride : int
        distance between the start of two consecutive window frames
    element_processor : function
        enables processing each non-discarded object; useful if ``stride >
        width`` and ``map_item`` is expensive (as compared to directly passing
        ``map(fun, itraj)`` as ``itraj``); if ``stride < width``, you could as
        well directly pass ``map(fun, itraj)``.
    """

    def __init__(self, itraj, width=2, stride=1, element_processor=None):

        self._raw_it = itraj
        if element_processor:
            self._it = map(element_processor, self._raw_it)
        else:
            self._it = self._raw_it
        assert(stride >= 1)
        assert(width >= 1)
        self.width = width
        self.stride = stride
        self._window = None

    def __iter__(self):
        return self

    def __next__(self):
        """ Returns next element in sequence. """
        if self._window is None:
            self._window = deque(islice(self._it, self.width), self.width)
        else:
            if self.stride >= self.width:
                self._window.clear()
                consume(self._raw_it, self.stride - self.width)
            else:
                for _ in range(min((self.stride, len(self._window)))):
                    self._window.popleft()
            for f in islice(self._it, min((self.stride, self.width))):
                self._window.append(f)

        if len(self._window) == 0:
            raise StopIteration

        return list(self._window)



# Let local_compiler be None in order to use the default compiler
local_compiler = None
local_linker = local_compiler
extra_compile_args = []
extra_link_args = []
local_link_shared = []

# # Example: Explicitly use gcc
# #
# # local_compiler = 'gcc'
# extra_compile_args = ['-fPIC',
#                       '-Ofast',
#                       '-march=native',
#                       '-std=c99']

# local_linker = local_compiler
# local_link_shared = ['-shared']
# extra_link_args = []


# # Example: Use icc instead of the default compiler
# #
# local_compiler = 'icc'
# extra_compile_args = ['-openmp', '-xHOST', '-O3',
#                       '-fno-alias', '-fPIC', '-std=c99']

# local_linker = local_compiler
# local_link_shared = ['-shared']
# extra_link_args = ['-openmp']


# # Example: Use pgcc instead of the default compiler
# #
# local_compiler = 'pgcc'
# extra_compile_args = ['-mp=numa', '-O4', '-Msafeptr', '-fPIC', '-c99']

# local_linker = local_compiler
# local_link_shared = ['-shared']
# extra_link_args = ['-mp']


# # Example: Use pgcc to generate GPU-code using OpenACC directives
# #
# local_compiler = 'pgcc'
# extra_compile_args = ['-acc', '-ta=nvidia', '-O4',
#                       '-g', '-Minfo=acc', '-fPIC']

# local_linker = local_compiler
# local_link_shared = ['-shared']
# extra_link_args = ['-acc']



from .lammpstrj_trajectory_reader import lammpstrj_trajectory_reader
from .molfile_trajectory_reader import molfile_trajectory_reader
from .xtc_trajectory_reader import xtc_trajectory_reader

available_readers = []

for reader in [lammpstrj_trajectory_reader,
               molfile_trajectory_reader,
               xtc_trajectory_reader]:
    if reader.available():
        available_readers.append(reader)



__all__ = ['fixed_bin_averager']

import numpy as np
import logging

logger = logging.getLogger('dynasor')


class fixed_bin_averager:

    """
    Class for averaging data sets of points (x,y) with a priori decided
    positions (x), over a pre-defined number of equally sized, linearely
    distriuted bins.

    This is used to get an average of y for a set of lineary spaced x values
    when a possibly large set of {y(x)} is given.

    The result will be an array containing, for each bin, the average y-value
    over the data points having its x-value within the respective bin x-range.
    """

    def __init__(self, x_max, x_bins, x_distances, x_min=0.0):
        assert x_max > x_min
        assert x_bins > 1

        self.delta_x = (x_max-x_min) / (x_bins-1)
        x_range = (x_min-self.delta_x/2, x_max+self.delta_x/2)
        bin_count, edges = np.histogram(x_distances,
                                        bins=x_bins,
                                        range=x_range)
        self.x_linspace = 0.5 * (edges[1:]+edges[:-1])

        m = np.nonzero(bin_count)
        self.bin_count = bin_count[m]
        self.x = self.x_linspace[m]
        self.input_length = len(x_distances)
        self.bins = len(self.x)
        if self.bins != x_bins:
            logger.info('Ignoring {} bins without coverage'
                        .format(x_bins-self.bins))

    def bin(self, y, axis=0):
        y = np.require(y)
        assert y.shape[axis] == self.input_length

        res_shape = list(y.shape)
        res_shape[axis] = self.bins
        result = np.zeros(res_shape)

        ci = 0
        ind_x = [slice(None)]*len(y.shape)
        ind_y = [slice(None)]*len(y.shape)
        for i, n in enumerate(self.bin_count):
            ind_x[axis] = i
            ind_y[axis] = slice(ci, ci+n)
            result[tuple(ind_x)] = np.mean(y[tuple(ind_y)], axis=axis)
            ci += n

        return result
